@@ -81,8 +81,12 @@ typedef struct nano_handle_s {
 #include <errno.h>
 #endif
 
+#ifndef R_NO_REMAP
 #define R_NO_REMAP
+#endif
+#ifndef STRICT_R_HEADERS
 #define STRICT_R_HEADERS
+#endif
 #include <R.h>
 #include <Rinternals.h>
 #include <Rversion.h>
@@ -91,8 +95,10 @@ typedef struct nano_handle_s {
 #define NANO_TAG(x) TAG(x)
 #define NANO_PTR(x) (void *) CAR(x)
 #define NANO_PROT(x) CDR(x)
+#define NANO_ENCLOS(x) CDR(x)
 #define NANO_SET_TAG(x, v) SET_TAG(x, v)
 #define NANO_SET_PROT(x, v) SETCDR(x, v)
+#define NANO_SET_ENCLOS(x, v) SETCDR(x, v)
 #define NANO_DATAPTR(x) (void *) DATAPTR_RO(x)
 #define NANO_VECTOR(x) ((const SEXP *) DATAPTR_RO(x))
 #define NANO_STRING(x) CHAR(*((const SEXP *) DATAPTR_RO(x)))
@@ -102,7 +108,7 @@ typedef struct nano_handle_s {
 
 #define ERROR_OUT(xc) Rf_error("%d | %s", xc, nng_strerror(xc))
 #define ERROR_RET(xc) { Rf_warning("%d | %s", xc, nng_strerror(xc)); return mk_error(xc); }
-#define NANONEXT_INIT_BUFSIZE 8192
+#define NANONEXT_INIT_BUFSIZE 4096
 #define NANONEXT_SERIAL_VER 3
 #define NANONEXT_SERIAL_THR 134217728
 #define NANONEXT_ERR_STRLEN 40
@@ -131,16 +137,6 @@ typedef union nano_opt_u {
   uint64_t u;
 } nano_opt;
 
-typedef struct nano_listener_s {
-  nng_listener list;
-  nng_tls_config *tls;
-} nano_listener;
-
-typedef struct nano_dialer_s {
-  nng_dialer dial;
-  nng_tls_config *tls;
-} nano_dialer;
-
 typedef struct nano_stream_s {
   nng_stream *stream;
   union {
@@ -170,12 +166,18 @@ typedef enum nano_aio_typ {
 typedef struct nano_aio_s {
   nng_aio *aio;
   void *data;
+  void *cb;
   void *next;
-  int mode;
   int result;
+  uint8_t mode;
   nano_aio_typ type;
 } nano_aio;
 
+typedef struct nano_saio_s {
+  nng_aio *aio;
+  void *cb;
+} nano_saio;
+
 typedef struct nano_cv_s {
   int condition;
   int flag;
@@ -200,109 +202,32 @@ typedef struct nano_thread_duo_s {
   nano_cv *cv2;
 } nano_thread_duo;
 
+typedef struct nano_thread_disp_s {
+  nng_thread *thr;
+  nano_cv *cv;
+  nng_tls_config *tls;
+  nano_saio **saio;
+  nano_aio **raio;
+  nano_aio **haio;
+  nng_url *up;
+  const char *host;
+  char **url;
+  int *online;
+  R_xlen_t n;
+} nano_thread_disp;
+
+typedef struct nano_signal_s {
+  nano_cv *cv;
+  int *online;
+} nano_signal;
+
 typedef struct nano_buf_s {
   unsigned char *buf;
   size_t len;
   size_t cur;
 } nano_buf;
 
-void dialer_finalizer(SEXP);
-void listener_finalizer(SEXP);
-void socket_finalizer(SEXP);
-SEXP R_mkClosure(SEXP, SEXP, SEXP);
-void later2(void (*)(void *), void *);
 extern void (*eln2)(void (*)(void *), void *, double, int);
-void eln2dummy(void (*)(void *), void *, double, int);
-int nano_integer(SEXP);
-SEXP mk_error(const int);
-SEXP mk_error_data(const int);
-SEXP rawToChar(const unsigned char *, const size_t);
-void nano_serialize_old(nano_buf *, const SEXP, SEXP);
-void nano_serialize(nano_buf *, const SEXP, SEXP);
-SEXP nano_unserialize(unsigned char *, const size_t, SEXP);
-SEXP nano_decode(unsigned char *, const size_t, const int, SEXP);
-void nano_encode(nano_buf *, const SEXP);
-int nano_encodes(const SEXP);
-int nano_matcharg(const SEXP);
-int nano_matchargs(const SEXP);
-SEXP nano_PreserveObject(SEXP);
-void nano_ReleaseObject(SEXP);
-
-SEXP rnng_advance_rng_state(void);
-SEXP rnng_aio_call(SEXP);
-SEXP rnng_aio_collect(SEXP);
-SEXP rnng_aio_collect_pipe(SEXP);
-SEXP rnng_aio_collect_safe(SEXP);
-SEXP rnng_aio_get_msg(SEXP);
-SEXP rnng_aio_http_data(SEXP);
-SEXP rnng_aio_http_headers(SEXP);
-SEXP rnng_aio_http_status(SEXP);
-SEXP rnng_aio_result(SEXP);
-SEXP rnng_aio_stop(SEXP);
-SEXP rnng_clock(void);
-SEXP rnng_close(SEXP);
-SEXP rnng_ctx_close(SEXP);
-SEXP rnng_ctx_create(SEXP);
-SEXP rnng_ctx_open(SEXP);
-SEXP rnng_cv_alloc(void);
-SEXP rnng_cv_reset(SEXP);
-SEXP rnng_cv_signal(SEXP);
-SEXP rnng_cv_until(SEXP, SEXP);
-SEXP rnng_cv_until_safe(SEXP, SEXP);
-SEXP rnng_cv_value(SEXP);
-SEXP rnng_cv_wait(SEXP);
-SEXP rnng_cv_wait_safe(SEXP);
-SEXP rnng_dial(SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_dialer_close(SEXP);
-SEXP rnng_dialer_start(SEXP, SEXP);
-SEXP rnng_fini(void);
-SEXP rnng_get_opt(SEXP, SEXP);
-SEXP rnng_is_error_value(SEXP);
-SEXP rnng_is_nul_byte(SEXP);
-SEXP rnng_listen(SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_listener_close(SEXP);
-SEXP rnng_listener_start(SEXP);
-SEXP rnng_messenger(SEXP);
-SEXP rnng_messenger_thread_create(SEXP);
-SEXP rnng_ncurl(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_ncurl_aio(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_ncurl_session(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_ncurl_session_close(SEXP);
-SEXP rnng_ncurl_transact(SEXP);
-SEXP rnng_next_config(SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_pipe_close(SEXP);
-SEXP rnng_pipe_notify(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_protocol_open(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_random(SEXP, SEXP);
-SEXP rnng_reap(SEXP);
-SEXP rnng_recv(SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_recv_aio(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_request(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_send(SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_send_aio(SEXP, SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_serial_config(SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_set_marker(SEXP);
-SEXP rnng_set_promise_context(SEXP, SEXP);
-SEXP rnng_set_opt(SEXP, SEXP, SEXP);
-SEXP rnng_signal_thread_create(SEXP, SEXP);
-SEXP rnng_sleep(SEXP);
-SEXP rnng_socket_lock(SEXP, SEXP);
-SEXP rnng_socket_unlock(SEXP);
-SEXP rnng_stats_get(SEXP, SEXP);
-SEXP rnng_status_code(SEXP);
-SEXP rnng_stream_close(SEXP);
-SEXP rnng_stream_dial(SEXP, SEXP, SEXP);
-SEXP rnng_stream_listen(SEXP, SEXP, SEXP);
-SEXP rnng_strerror(SEXP);
-SEXP rnng_subscribe(SEXP, SEXP, SEXP);
-SEXP rnng_tls_config(SEXP, SEXP, SEXP, SEXP);
-SEXP rnng_unresolved(SEXP);
-SEXP rnng_unresolved2(SEXP);
-SEXP rnng_url_parse(SEXP);
-SEXP rnng_version(void);
-SEXP rnng_wait_thread_create(SEXP);
-SEXP rnng_write_cert(SEXP, SEXP, SEXP);
-
 extern uint8_t special_bit;
 
 extern SEXP nano_AioSymbol;
@@ -331,7 +256,6 @@ extern SEXP nano_aioFuncMsg;
 extern SEXP nano_aioFuncRes;
 extern SEXP nano_aioNFuncs;
 extern SEXP nano_error;
-extern SEXP nano_klassString;
 extern SEXP nano_precious;
 extern SEXP nano_recvAio;
 extern SEXP nano_reqAio;
@@ -339,4 +263,113 @@ extern SEXP nano_sendAio;
 extern SEXP nano_success;
 extern SEXP nano_unresolved;
 
+#if R_VERSION < R_Version(4, 1, 0)
+SEXP R_NewEnv(SEXP, int, int);
+#endif
+#if R_VERSION < R_Version(4, 5, 0)
+SEXP R_mkClosure(SEXP, SEXP, SEXP);
+#endif
+SEXP nano_findVarInFrame(const SEXP, const SEXP);
+SEXP nano_PreserveObject(const SEXP);
+void nano_ReleaseObject(SEXP);
+void raio_complete_signal(void *);
+void sendaio_complete(void *);
+void cv_finalizer(SEXP);
+void dialer_finalizer(SEXP);
+void listener_finalizer(SEXP);
+void socket_finalizer(SEXP);
+void later2(void (*)(void *), void *);
+void raio_invoke_cb(void *);
+int nano_integer(const SEXP);
+SEXP mk_error(const int);
+SEXP mk_error_data(const int);
+SEXP rawToChar(const unsigned char *, const size_t);
+void nano_serialize(nano_buf *, const SEXP, SEXP);
+SEXP nano_unserialize(unsigned char *, const size_t, SEXP);
+SEXP nano_decode(unsigned char *, const size_t, const uint8_t, SEXP);
+void nano_encode(nano_buf *, const SEXP);
+int nano_encodes(const SEXP);
+int nano_matcharg(const SEXP);
+int nano_matchargs(const SEXP);
+
+void pipe_cb_signal(nng_pipe, nng_pipe_ev, void *);
+void tls_finalizer(SEXP);
+
+SEXP rnng_advance_rng_state(void);
+SEXP rnng_aio_call(SEXP);
+SEXP rnng_aio_collect(SEXP);
+SEXP rnng_aio_collect_pipe(SEXP);
+SEXP rnng_aio_collect_safe(SEXP);
+SEXP rnng_aio_get_msg(SEXP);
+SEXP rnng_aio_http_data(SEXP);
+SEXP rnng_aio_http_headers(SEXP);
+SEXP rnng_aio_http_status(SEXP);
+SEXP rnng_aio_result(SEXP);
+SEXP rnng_aio_stop(SEXP);
+SEXP rnng_clock(void);
+SEXP rnng_close(SEXP);
+SEXP rnng_ctx_close(SEXP);
+SEXP rnng_ctx_create(SEXP);
+SEXP rnng_ctx_open(SEXP);
+SEXP rnng_cv_alloc(void);
+SEXP rnng_cv_reset(SEXP);
+SEXP rnng_cv_signal(SEXP);
+SEXP rnng_cv_until(SEXP, SEXP);
+SEXP rnng_cv_until_safe(SEXP, SEXP);
+SEXP rnng_cv_value(SEXP);
+SEXP rnng_cv_wait(SEXP);
+SEXP rnng_cv_wait_safe(SEXP);
+SEXP rnng_dial(SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_dialer_close(SEXP);
+SEXP rnng_dialer_start(SEXP, SEXP);
+SEXP rnng_dispatcher_socket(SEXP, SEXP, SEXP);
+SEXP rnng_eval_safe(SEXP);
+SEXP rnng_fini(void);
+SEXP rnng_get_opt(SEXP, SEXP);
+SEXP rnng_is_error_value(SEXP);
+SEXP rnng_is_nul_byte(SEXP);
+SEXP rnng_listen(SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_listener_close(SEXP);
+SEXP rnng_listener_start(SEXP);
+SEXP rnng_messenger(SEXP);
+SEXP rnng_messenger_thread_create(SEXP);
+SEXP rnng_ncurl(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_ncurl_aio(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_ncurl_session(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_ncurl_session_close(SEXP);
+SEXP rnng_ncurl_transact(SEXP);
+SEXP rnng_pipe_close(SEXP);
+SEXP rnng_pipe_notify(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_protocol_open(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_random(SEXP, SEXP);
+SEXP rnng_read_online(SEXP);
+SEXP rnng_reap(SEXP);
+SEXP rnng_recv(SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_recv_aio(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_request(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_send(SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_send_aio(SEXP, SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_serial_config(SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_set_marker(SEXP);
+SEXP rnng_set_opt(SEXP, SEXP, SEXP);
+SEXP rnng_set_promise_context(SEXP, SEXP);
+SEXP rnng_signal_thread_create(SEXP, SEXP);
+SEXP rnng_sleep(SEXP);
+SEXP rnng_socket_lock(SEXP, SEXP);
+SEXP rnng_socket_unlock(SEXP);
+SEXP rnng_stats_get(SEXP, SEXP);
+SEXP rnng_status_code(SEXP);
+SEXP rnng_stream_close(SEXP);
+SEXP rnng_stream_dial(SEXP, SEXP, SEXP);
+SEXP rnng_stream_listen(SEXP, SEXP, SEXP);
+SEXP rnng_strerror(SEXP);
+SEXP rnng_subscribe(SEXP, SEXP, SEXP);
+SEXP rnng_tls_config(SEXP, SEXP, SEXP, SEXP);
+SEXP rnng_unresolved(SEXP);
+SEXP rnng_unresolved2(SEXP);
+SEXP rnng_url_parse(SEXP);
+SEXP rnng_version(void);
+SEXP rnng_wait_thread_create(SEXP);
+SEXP rnng_write_cert(SEXP, SEXP, SEXP);
+
 #endif
